## get_osc.py
#
#  Get OSC data and decode
#
#############################################################################

import GameLogic as gl

import socket

from OSC3.OSC import decodeOSC

from time import sleep

scenes = gl.getSceneList()

def main():
    ip = gl.ip_multicast
    buffer_size = 1024
    port = gl.port_in

    connect_blender(ip, port, buffer_size)
    get_OSC_data(buffer_size)

    # Calcul et affichage du ratio de réception, le socket ne récupère pas toujours des datas à chaque frame
    gl.frame_nbr += 1
    if gl.frame_nbr == 1000:
        gl.frame_nbr, gl.socket_failed = 0, 0
    if gl.frame_nbr != 0:
        gl.socket_ratio = (gl.frame_nbr - gl.socket_failed)/gl.frame_nbr

def connect_blender(ip, port, buffer_size):

    # Connexion de Blender en Multicast
    if not gl.connected :
        gl.connected = True
        ANY = "0.0.0.0"
        MCAST_ADDR = "224.0.0.14"
        MCAST_PORT = 18888

        #create a UDP socket
        gl.sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM, socket.IPPROTO_UDP)
        #allow multiple sockets to use the same PORT number
        gl.sock.setsockopt(socket.SOL_SOCKET,socket.SO_REUSEADDR,1)
        #Bind to the port that we know will receive multicast data
        gl.sock.bind((ANY,MCAST_PORT))
        #tell the kernel that we are a multicast socket
        gl.sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 255)
        #Tell the kernel that we want to add ourselves to a multicast group
        #The address for the multicast group is the third param
        status = gl.sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, socket.inet_aton(MCAST_ADDR) + socket.inet_aton(ANY))
        gl.sock.setblocking(0)
        # Limite la taille du buffer UDP pour éviter la latence, le buffer est vidé à chaque lecture
        gl.sock.setsockopt(socket.SOL_SOCKET, socket.SO_RCVBUF, buffer_size)

def get_OSC_data(buffer_size):
    # Si blender est connecté, récupération des messages OSC
    if gl.connected :
        try:
            raw_data = gl.sock.recv(buffer_size)
            data     = decodeOSC(raw_data)
            #print("data in get_osc.py =", data)
            #if gl.tempo_globale == 0:
                #print("data in get_osc.py =", data)
            if '#bundle' in data:
                recup_messages(data)
            elif '/ip':
                print("IP server = ", data[2])
                gl.ip_server = data[2]
        except socket.error:
            gl.socket_failed += 1
            pass

def recup_messages(data):
    ''' Pour 1 joueur
     data = ['#bundle', 0.0,
            ['/fap', ',siiiiisssi','192.168.1.4', -374, 0, -950, 0, 9, 'serge', 'classement', 'serge', 0],
            ['/f', ',i', 0],
            ['/rbs', ',i', 0],
            ['/grs', ',i', 0],
            ['/qs', ',s', ''],
            ['/sr', ',i', 0]]
    '''
    # Cut '#bundle', 0.0
    data = data[2:]
    # récup des messages
    for item in data:
        if item[0] == '/fap':
            # Cette fonction a été créèe après le script players_data.py qui utilise gl.data
            # pour mettre en forme gl.players_data
            gl.data = data[0]
        elif item[0] == '/f':
            # item = ['/f', ',s', 'j2']
            if item[2] != 0:
                final_display(winner=item[2])
            else:
                gl.rank_scene = 0
                gl.text_rank = 0
        elif item[0] == "/rbs":
            if item[2] == 1:
                ball_reset()
        elif item[0] == '/grs':
            if item[2] == 1:
                general_reset()
        elif item[0] == '/qs':
            if item[2] != "":
               quit_game(item[2])
        elif item[0] == '/sr':
            if item[2] == 1:
                reset_score()

def final_display(winner):
    # Lancé si le message '/f' = winner et non False
    print("OSC message /f received")
    # Affichage de la scène Rank
    if gl.level > 1:
        # Si Rank n'est pas encore affiché
        if gl.rank_scene != 2:
            # Provoque l'affichage de la scène Rank
            gl.rank_scene = 1
    # Affichage en bas d'écran de ce qui se passe
    gl.text_rank = 1

def ball_reset():
    # Lancé si le message '/ball_reset' = 1
    print("OSC message /rbs received: balle au centre")
    gl.ball.localPosition = [3, 3, 1]

def general_reset():
    # Lancé si le message '/general_reset' = 1
    print("OSC message /grs: new_game(), reset_variables(), reset_score()")
    new_game()

def quit_game(joueur):
    # Reçu si un joueur quitte
    print("Un joueur a quitté le jeu:", joueur)

def reset_score():
    # Reset des scores du niveau en cours
    print ("OSC message /sr received: Je reset les scores")
    for i in range(gl.level):
        gl.goal[i]["score"] = 9

def reset_variables():
    # Reset de variables pour repartir à zéro
    print ("OSC message /sr received: Je reset les variables")
    reset_score()
    gl.rank_display_bug = 0
    gl.tempo_globale = 0

    ### level 1 only
    ##gl.level1_rated = 0 # si classement du level 1 fait dans game.py
    ##gl.classement_level1 = {}
    ##gl.tempo_rank_level1 = 0

    # pour afficher le classement si j'ai perdu
    gl.rank_scene = 0

    # Le dictionnaire de classement des joueurs
    gl.classement = {}

    # blocage de la bat si j'ai fini
    gl.block = 0

def new_game():
    # Relance tout à zéro
    print("OSC message /grs received: new_game()" )
    reset_variables()
    ### Supprime la scène précédente, ajoute la scene 1
    for scn in scenes :
        if "Rank" in scn.name:
            scn.end()
        if "_players" in scn.name:
            scn.end()

cont = gl.getCurrentController()
sens = cont.sensors["Get OSC"]
if sens.positive:
    main()
