#!/usr/bin/env python
# -*- coding: UTF-8 -*-

#############################################################################
# Copyright (C) Labomedia May 2012
#
#  This program is free software; you can redistribute it and/or
#  modify it under the terms of the GNU General Public License
#  as published by the Free Software Foundation; either version 2
#  of the License, or (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software Foundation,
#  Inc., 51 Franproplin Street, Fifth Floor, Boston, MA 02110-1301, USA.
#
#############################################################################

import os
import sys
import threading
import subprocess
from time import time, sleep

from txosc import async, dispatch, osc
from twisted.internet import reactor

import managementDict

class UDPMulticastSender(object):
    def __init__(self, port, host):
        self.port = port
        self.host = host
        self.client = async.DatagramClientProtocol()
        self._client_port = reactor.listenUDP(0, self.client)
        print("Sending multicast on: %s : %s" % (self.host, self.port))

    def send_message(self, message):
        self.client.send(message, (self.host, self.port))

class UDPReceiverApplication(object):
    def __init__(self, port):
        self.port = port
        self.receiver = dispatch.Receiver()
        self._server_port = reactor.listenUDP(self.port, async.DatagramServerProtocol(self.receiver))
        print("Listening on osc.udp://localhost:%s" % (self.port))
        # handler
        self.receiver.addCallback("/d", ServerMain_objet.player_data_handler) #"/player_data"
        self.receiver.addCallback("/rb", ServerMain_objet.reset_ball_handler) # "/reset_ball"
        self.receiver.addCallback("/rg", ServerMain_objet.reset_game_handler) # "/reset_game"
        self.receiver.addCallback("/q", ServerMain_objet.quit_handler) # "/quit"

class ServerMain(object):
    def __init__(self, port_in, port_multicast_out,host_multicast):
        self.port_in = port_in
        self.port_multicast_out = port_multicast_out
        self.host_multicast = host_multicast

        # Dictionnaire de centralisation des messages et des joueurs
        self.gameDict = managementDict.GameManagement()
        self.winner = 0
        self.since = 0

        # Fréquence d' ACTA
        self.t_connected = time()

        # messages particuliers
        self.msg_reset_ball = osc.Message("/rbs")
        self.msg_reset_ball.add(0)
        self.msg_new = osc.Message("/grs")
        self.msg_new.add(0)
        self.msg_final = osc.Message("/f")
        self.msg_final.add(0)
        self.msg_quit_svr = osc.Message("/qs")
        self.msg_quit_svr.add("")
        self.msg_score_reset = osc.Message("/sr")
        self.msg_score_reset.add(0)

    def never_sleeping(self):
        # Envoi toutes les secondes de l'adresse ip du serveur sur le Multicast
        while True:
            sleep(1)
            # Send message with Server IP
            msg_ip = osc.Message("/ip")
            msg_ip.add(my_ip)
            # Envoi
            self.multicast_sender.send_message(msg_ip)
            print "Je ne dors pas, j'envoie l'adresse IP"

    def game_server(self):
        # Instance de twisted avec txosc
        self.multicast_sender = UDPMulticastSender(self.port_multicast_out, self.host_multicast)
        self.UDP_receiver = UDPReceiverApplication(self.port_in)
        print "Serveur du jeu MultiPass For Freedom\n Press q + Enter to quit"
        print "Je lance Reactor"
        reactor.run()

    def player_data_handler(self, message, address):
        '''
        Cette fonction est appelée à chaque réception de message.
        Ajout des datas à gameDict, gestion des déconnectés, gestion du gagnant
        '''
        # Ajout à gameDict
        self.gameDict.insert_data_in_dict(message.getValues(), address)

        # Gestion des connectés
        self.ACTA()

        # The winner
        self.somebody_win()

    def ACTA(self):
        # Gestion des déconnectés toutes les secondes
        if time() - self.t_connected > 1:
            self.t_connected = time()
            # Gestion des déconnectés et connectés
            connected = self.gameDict.delete_disconnected_player()
            print "Les connectés sont:", connected

    def send2players(self):
        '''La fréquende d'envoi doit être un peu supérieure à 60 Hz, le FPS de Blender est au mieux de 60'''
        while True:
            sleep(0.02)
            # Création d'un bundle qui contiendra tous les messages
            b = self.create_bundle()
            # Envoi
            self.multicast_sender.send_message(b)

    def create_bundle(self):
        # Création d'un bundle qui contiendra tous les messages
        b = osc.Bundle()

        # Message du dictionnaire gameDict
        msg_for_all_players = self.gameDict.create_msg_for_all_players()
        b.add(msg_for_all_players)
        # Message du winner
        b.add(self.msg_final)

        # Message reset ball
        b.add(self.msg_reset_ball)
        # Remise à zéro
        self.msg_reset_ball = self.reinit_message("/rbs", 0)

        # Message reset general
        b.add(self.msg_new)
        # Remise à zéro
        self.msg_new = self.reinit_message("/grs", 0)

        # Message quit_svr
        b.add(self.msg_quit_svr)
        # Remise à zéro
        self.msg_quit_svr = self.reinit_message("/qs", "")

        # Message score_reset
        b.add(self.msg_score_reset)
        # Remise à zéro
        self.msg_score_reset = self.reinit_message("/sr", 0)

        return b

    def reinit_message(self, msg, value):
        # Réinitiale un message après envoi
        msg_reinit = osc.Message(msg)
        msg_reinit.add(value)
        return msg_reinit

    def reset_ball_handler(self, message, address):
        print ("/rb received")
        self.msg_reset_ball = osc.Message("/rbs")
        self.msg_reset_ball.add(1)

    def reset_game_handler(self, message, address):
        print ("/rg received")
        # Send message /general_reset
        self.msg_new = osc.Message("/grs")
        self.msg_new.add(1)
        self.gameDict.reset_data()

    def somebody_win(self):
        # Cette fonction est lancée à chaque réception, donc 120 Hz avec 2 joueurs
        self.winner = self.gameDict.who_is_the_winner()
        self.since = self.gameDict.winner_since()

        self.msg_final = self.reinit_message("/f", self.winner)

        if self.since > 1.5:
            # Remise à 9 de tous les scores de tous les joueurs
            # Ce message doit être envoyé plusieurs fois
            self.msg_score_reset = self.reinit_message("/sr", 1)

    def quit_handler(self, message, address):
        # Only to display in blender terminal
        print "Quit received from:", address
        self.msg_quit_svr = osc.Message("/rbs")
        self.msg_quit_svr.add(0)

def thread_keyboard():
    '''doc'''
    while True:
        touche = sys.stdin.read(1)
        if touche == 'q':
            print("Exit")
            os._exit(0)

# -------- Get local ip -------#
if os.name != "nt":
    import socket
    import fcntl
    import struct

    def get_interface_ip(ifname):
        s = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        return socket.inet_ntoa(fcntl.ioctl(s.fileno(), 0x8915, struct.pack('256s',
                                ifname[:15]))[20:24])

    def get_lan_ip():
        ip = socket.gethostbyname(socket.gethostname())
        if ip.startswith("127.") and os.name != "nt":
            interfaces = [
                "eth0",
                "eth1",
                "eth2",
                "wlan0",
                "wlan1",
                "wifi0",
                "ath0",
                "ath1",
                "ppp0",
                ]
            for ifname in interfaces:
                try:
                    ip = get_interface_ip(ifname)
                    break
                except IOError:
                    pass
        return ip

if __name__ == "__main__":
    host_multicast = '224.0.0.14'
    port_multicast_out = 18888
    port_in  = 19999

    # Mon ip à envoyer au joueur
    my_ip = get_lan_ip()
    print "Mon IP =", my_ip

    # Attente d'une saisie clavier
    thread1 = threading.Thread(target=thread_keyboard)
    thread1.start()

    # set instance of ServerMain class
    ServerMain_objet = ServerMain(port_in, port_multicast_out, host_multicast)

    # Envoi des bundles
    thread2 = threading.Thread(target=ServerMain_objet.send2players)
    thread2.start()

    print "Le serveur est lancé"

    # Envoi des bundles
    thread3 = threading.Thread(target=ServerMain_objet.never_sleeping)
    thread3.start()

    # Run server
    ServerMain_objet.game_server()
    # La fonction game_server() se termine par reactor.run() qui lance l'envoi et la réception de twisted,
    # l'exécution de cette fonction s'arrête le script ici.
    print "Je ne passe jamais ici"
